from unittest import TestCase
from requests.models import Response
from src.access_logs_local_driver.process_download_logs import (
    make_filters,
    no_plus_http,
    no_star,
)


class TestProcessDownLogs(TestCase):
    def setUp(self) -> None:
        self.modes = [
            {
                "measure": "https://test.test-eu.org/abc/downloads/v1",
                "name": "download",
                "regex": [
                    (
                        "https://abcdef.hijkl.info/articles/10.\\d{4,9}"
                        "/[-._;()/:a-zA-Z0-9]+/galley/\\d+/download"
                    )
                ],
            },
            {
                "measure": "tag:test.eu,2018:test:abc-html",
                "name": "htmlreader",
                "regex": [
                    (
                        "https://abcdef.hijkl.info/articles/10.\\d{4,9}/"
                        "[-._;()/:a-zA-Z0-9]+"
                    ),
                    (
                        "https://abcdef.hijkl.info/articles/abstract/10.\\"
                        "d{4,9}/[-._;()/:a-zA-Z0-9]+"
                    ),
                ],
            },
        ]
        self.spiders = (
            "user-agent=Mozilla/3.01Gold",
            "versus 0.2 (+http://testing.test.ch)",
            "versus crawler eda.testing@epfl.ch",
            "testing.nl - testing Bot/x.x",
            "testing/0.2",
            "testing",
            "test12345-hc/1.0",
            "test12345/1.0",
            "test12345/2.0 (http://www.testing.com/crawler.html)",
            "test12345/2.0 (http://www.testing.com/html/crawler.html)",
            "vspider",
            "vspider/3.x",
            "wadaino.jp-crawler 0.2 (http://testing.jp/)",
            "web-test12345 (Version: 1.02, powered by www.testing-test.de)",
            "web-test12345 (Version: 1.02, powered by www.web-test12345.de)",
            (
                "testing.org/test12345-0.9-dev (leveled playing field;"
                "http://test.org/; info at web2test.org)"
            ),
        )
        self.excluded_ips = '["8.8.8.8","9.9.9.9"]'
        self.excluded_urls = '["https://malformed_url.com"]'

    def test_response_has_no_star(self):
        the_response = Response()
        the_response.url = "nice_url/"
        self.assertTrue(no_star(the_response))
        the_response.url = "*"
        self.assertFalse(no_star(the_response))

    def test_response_no_plus_http(self):
        the_response = Response()
        the_response.user_agent = "http"
        self.assertTrue(no_plus_http(the_response))
        the_response.user_agent = "+http"
        self.assertFalse(no_plus_http(the_response))

    def test_make_filters(self):
        """Check the make_filters method is successful with the
        modes sample and returns the deisred methods."""
        for mode in self.modes:
            filters = (
                make_filters(
                    mode["regex"], ["8.8.8.8", "9.9.9.9"], self.spiders
                ),
                mode["regex"],
            )
        self.assertTrue(len(filters) > 0)
        self.assertIn(no_star, filters[0])
        self.assertIn(no_plus_http, filters[0])


    # def test_only_successful_response(self):
    #     """Returns true for only 200 and 304 responses"""
    #     the_response = Response()
    #     the_response.response_code = 200
    #     self.assertTrue(only_successful(the_response))
    #     the_response.response_code = 304
    #     self.assertTrue(only_successful(the_response))
    #     the_response.response_code = 404
    #     self.assertFalse(only_successful(the_response))
    #     the_response.response_code = 302
    #     self.assertFalse(only_successful(the_response))



import re
from typing import List, Callable, Optional

from .logdata import Request


def no_star(request: Request) -> bool:
    return request.url != "*"


def no_plus_http(request: Request) -> bool:
    return "+http" not in request.user_agent


def make_filters(
        regexes: List[str],
        excluded: List[str],
        allowed_codes: Optional[List[int]] = None
) -> List[Callable[[Request], bool]]:

    compiled_patterns = [re.compile(r) for r in regexes]
    target_codes = set(allowed_codes) if allowed_codes else {200, 304}
    excluded_ip_set = set(excluded)

    def filter_url(request):
        for pattern in compiled_patterns:
            if pattern.search(request.url):
                return True
        return False

    def only_successful(request: Request) -> bool:
        return request.response_code in target_codes

    def not_excluded_ip(request):
        return request.ip_address not in excluded_ip_set

    return [
        filter_url,
        only_successful,
        no_star,
        no_plus_http,
        not_excluded_ip,
    ]



import datetime
from sqlite3 import dbapi2 as sqlite


class GeoLookup:
    def __init__(self, path: str) -> None:
        if path is None:
            self.db = None
            return
        self.db = sqlite.connect(path)
        self.cursor = self.db.cursor()
        self.cache = {}
        self.query = '''
          SELECT country, datestamp FROM ipgeo
            WHERE ip_address = ?
            AND datestamp >= ?
            AND datestamp <= ?;
        '''
        self.prefix = 'urn:iso:std:3166:-2:'

    def lookup_country(self, ip_address: str, date: datetime) -> str:
        if self.db is None:
            return ''
        time_now = date.timestamp()
        time_max = time_now + 86400 * 180
        time_min = time_max - 86400 * 180
        args = (ip_address, time_min, time_max)
        self.cursor.execute(self.query, args)
        row = self.cursor.fetchone()
        if row is None:
            return ''
        return self.prefix + row[0]



from datetime import date, datetime
import ipaddress
from logging import getLogger
import re
import sys
from typing import Iterator, IO

import urllib.error
import urllib.parse


logger = getLogger(__name__)

LOG_REGEXES = {
    "apache": (
        r"(?P<ip_address>\d+\.\d+\.\d+\.\d+) "
        r"(?P<users>.+ .+) "
        r"\[(?P<timestamp>.+)\] "
        r'"(?P<request>.+)" '
        r"(?P<status_code>\d+) "        
        r"(?P<sent_bytes>\d+) "         
        r'(?P<referer>".+") '
        r'"(?P<user_agent>.+)"'
    ),
    "aws_alb": (  # AWS Application Load Balancer
        r"[^ ]+ "                           # type (e.g. h2)
        r"(?P<timestamp>[^ ]+) "            
        r"[^ ]+ "                           # elb_id
        r"(?P<ip_address>[^:]+):[^ ]+ "     # client:port (Capture IP only)
        r"[^ ]+ "                           # target:port
        r"[^ ]+ [^ ]+ [^ ]+ "               # processing times
        r"(?P<status_code>\d+) "            # elb_status_code
        r"[^ ]+ [^ ]+ "                     # target_status_code received_bytes
        r"(?P<sent_bytes>\d+) "             
        r"\"(?P<request>[^\"]+)\" "         
        r"\"(?P<user_agent>[^\"]+)\" "      
        r".*"                               # Ignore the rest (ssl, arn, etc)
    )
}


class Request:
    """Represent the data in a single line of the Apache log file."""

    def __init__(
            self,
            ip_address: str,
            url_prefix: str,
            url: str,
            response_code: int,
            content_length: int,
            timestamp: datetime,
            user_agent: str,
    ):
        self.ip_address = ip_address
        self.timestamp = timestamp
        self.user_agent = user_agent or ""
        self.url = self.parse_url(url, url_prefix)
        self.response_code = response_code
        self.content_length = content_length

    def parse_url(self, url: str, url_prefix: str) -> str:
        try:
            if url.startswith("http"):
                return url_prefix + urllib.parse.urlparse(url).path.lower()
            return self.normalise_url(url_prefix + url.lower())
        except ValueError:
            raise ValueError(f"Error parsing: {url}, {sys.stderr}")

    @staticmethod
    def normalise_url(url: str) -> str:
        try:
            return url[:-1] if url[-1] == "/" else url
        except IndexError as err:
            raise IndexError(f"Error parsing: {url}, {err}")

    def fmttime(self) -> str:
        return self.timestamp.strftime("%Y-%m-%d %H:%M:%S")

    def __str__(self) -> str:
        return f"Request {self.fmttime()}, {self.ip_address}, {self.url}"

    def __iter__(self):
        for _item in self.as_tuple():
            yield _item

    def as_tuple(self) -> tuple[str, str, str, str]:
        return self.fmttime(), self.ip_address, self.url, self.user_agent

    def sanitise_url(self, regexes: str) -> None:
        for regex in regexes:
            matched = re.search(re.compile(regex), self.url)
            if matched is not None:
                self.url = matched.group(0)
                break


class LogProcessor:
    """Consumes an open file stream, parses lines, applies filters,
    and yields Request objects.
    """

    def __init__(
            self,
            filter_groups: list,
            url_prefix: str,
            start_date: str,
            end_date: str,
            log_type: str = "apache",
            log_re_pattern: str | None = None,
    ) -> None:
        self.filter_groups = filter_groups
        self.url_prefix = url_prefix
        self.start_date = date.fromisoformat(start_date)
        self.end_date = date.fromisoformat(end_date)

        self.log_regex_patterns = LOG_REGEXES
        self.log_type = log_type
        self._set_log_regex(log_type, log_re_pattern)

    def _set_log_regex(self, log_type: str, log_pattern) -> None:
        """Attempt to determine regex pattern to use based on log type."""
        if log_type is None:
            self.log_regex = None
            return None
        elif log_type in LOG_REGEXES:
            log_pattern = LOG_REGEXES[log_type]
        elif log_type == "custom":
            if not log_pattern:
                ValueError(
                    f"`log_re_pattern` must be specified for "
                    f"`log_type` value 'custom'."
                )

        self.log_regex = re.compile(log_pattern)
        return None

    def _determine_log_regex(self, line: str) -> tuple:
        """Determine the regex pattern that matches the log line. If no pattern
        matches, raises a ValueError.

        Args:
            line (str): Log line to match against the registered regex patterns.

        Raises:
            ValueError: If no regex pattern matches the log line.

        Returns:
            tuple: The regex pattern that matches the given log line.
        """
        for log_type, log_pattern in self.log_regex_patterns.items():
            match = re.search(log_pattern, line)
            if match:
                return log_type, log_pattern
        else:
            raise ValueError(
                f"No regex patterns found that match the format of the log file. "
                f"Please set a custom regex pattern."
            )

    def process(self, file_stream: IO[str]) -> Iterator[tuple]:
        """Takes a file-like object (text mode), iterates it, and yields
         filtered results.
        """
        for line in file_stream:
            if isinstance(line, bytes):  # Handle mixed bytes/string input.
                line = line.decode("utf-8")

            if line := line.strip():
                request_obj = self._parse_line(line)
                if request_obj:
                    for filter_group in self.filter_groups:
                        measure_uri, filters, regex = filter_group

                        if self._apply_filters(filters, request_obj):
                            request_obj.sanitise_url(regex)
                            yield measure_uri, request_obj
                            break  # only allow one measure to match per line

    def _parse_line(self, line: str) -> Request | None:
        """Internal method to match regex and create a Request object"""
        if not self.log_regex:
            self.log_type, log_pattern = self._determine_log_regex(line)
            self._set_log_regex(self.log_type, log_pattern)

        match = self.log_regex.search(line)
        data = match.groupdict()

        try:
            if self.log_type == "apache":
                timestamp = datetime.strptime(
                    data.pop("timestamp"),
                    "%d/%b/%Y:%H:%M:%S %z"
                )
            elif self.log_type == "aws_alb":
                ts_str = data.pop("timestamp")
                timestamp = datetime.fromisoformat(ts_str)
            else:
                logger.error(
                    f"Log type {self.log_type} is not currently supported."
                )  # TODO: Add support for custom timestamp format.
                raise ValueError("Invalid log type")
        except ValueError:
            return None

        if not self.start_date <= timestamp.date() <= self.end_date:
            return None

        user_agent = data.get("user_agent", "-")
        ip = data.pop("ip_address")
        if not self._validate_ip(ip):
            return None

        if url := self._validate_request(data.pop("request")):
            return Request(
                ip_address=ip,
                url_prefix=self.url_prefix,
                url=url,
                response_code=int(data["status_code"]),
                content_length=int(data["sent_bytes"]),
                timestamp=timestamp,
                user_agent=user_agent,
            )

    @staticmethod
    def _apply_filters(filters, request_obj) -> bool:
        return all(f(request_obj) for f in filters)

    @staticmethod
    def _validate_request(request_str):
        parts = request_str.split()
        if len(parts) == 3:
            return parts[1]
        return None

    @staticmethod
    def _validate_ip(ip_address_str):
        """Validates whether the string is a valid IPv4 or IPv6 address."""
        try:
            ipaddress.ip_address(ip_address_str)
            return True
        except ValueError:
            return False  # TODO: fix country code for IP6 Addresses



import gzip
import os
import pathlib as pl
from unittest import TestCase

from src.access_logs_local_driver import make_filters, LogProcessor


class TestLogData(TestCase):
    def setUp(self) -> None:
        self.modes_cjs = [
            {
                "measure": "https://testing.test-eu.org/abc/downloads/v1",
                "name": "download",
                "regex": [
                    (
                        "https://abc.test.testing/testing-url/10.\\d{4,9}/"
                        "[-._;()/:a-zA-Z0-9]+/galley/\\d+/download"
                    )
                ],
            },
            {
                "measure": "tag:testing.eu,2023:readership:abc-html",
                "name": "htmlreader",
                "regex": [
                    (
                        "https://abc.test.testing/testing-url/10.\\d{4,9}/"
                        "[-._;()/:a-zA-Z0-9]+"
                    ),
                    (
                        "https://abc.test.testing/testing-url/abstract/10.\\d"
                        "{4,9}/[-._;()/:a-zA-Z0-9]+"
                    ),
                ],
            },
        ]
        self.modes_nginx = [
            {
                "measure": "https://testing.test-eu.org/abc/v1",
                "name": "download",
                "regex": [
                    "https://abc.test.testing/test" "[-._;()/:a-zA-Z0-9]+"
                ],
            },
        ]
        self.file_log_apache = "test_access.log-20230602.gz"
        self.file_log_nginx = "test_access.log-20230705.gz"
        self.cache_dir = "tests/files/"
        self.logs_files = self.cache_dir
        self.url_prefix = "https://abc.test.testing"

    def tearDown(self):
        """Check whether the file has been created by any test and
        delete them, useful when running tests separately."""
        if self.file_exists(self.logs_files + self.file_log_apache):
            os.remove(self.logs_files + self.file_log_apache)
        if self.file_exists(self.logs_files + "test_access.log-NODATE.gz"):
            os.remove(self.logs_files + "test_access.log-NODATE.gz")
        if self.file_exists(self.logs_files + self.file_log_nginx):
            os.remove(self.logs_files + self.file_log_nginx)

    def create_file_log(self, file_name, content):
        file = gzip.open(self.logs_files + file_name, "wb")
        file.write(content)
        file.close()

    def file_exists(self, path):
        if not pl.Path(path).resolve().is_file():
            return False
        return True

    def test_apache_log_stream_creates_is_succesful_one_match_cjs(self) -> None:
        """Test the class LogStream the same way as is executed
        from the plugin"""
        start_date = '2023-06-01'
        end_date = '2023-06-03'
        content = str.encode(
            "cjs.testing-url.com:123 12.90.253.37 - - "
            '[02/Jun/2023:04:55:37 +0000] "GET /testing-url/'
            "10.3456768/cjs.abcde.1234/image/1234/download/ "
            'HTTP/1.1" 200 123456 "https://www.google.com/" '
            '"Mozilla/5.0 (Linux; Android 13; testPhone ABCDEF) '
            "TestPhone/123.45 (KHTML, like Gecko) TestBrowser"
            '/21.0 Chrome/120.3.4567.890 Mobile Safari/537.36"'
        )
        self.create_file_log(self.file_log_apache, content)
        filter_groups = []
        for mode in self.modes_cjs:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)
        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with gzip.open(
                self.logs_files + self.file_log_apache,
                "rt",
                encoding="utf-8"
        ) as f:
            result = list(processor.process(f))

        # Assert the result has got the right content
        self.assertEqual(
            result[0][0],
            'tag:testing.eu,2023:readership:abc-html',
        )
        self.assertEqual(
            str(result[0][1]),
            (
                "Request 2023-06-02 04:55:37, 12.90.253.37, "
                "https://abc.test.testing/testing-url/10.3456768"
                "/cjs.abcde.1234/image/1234/download"
            )
        )

    def test_apache_log_stream_creates_is_succesful_two_matches_cjs(
        self
    ) -> None:
        """Test the class LogStream the same way as is executed
        from the plugin"""
        start_date = '2023-06-01'
        end_date = '2023-06-02'
        content = str.encode(
            "cjs.testing-url.com:123 12.34.253.12 - - "
            '[02/Jun/2023:04:55:37 +0000] "GET /testing-url/'
            "10.3456768/cjs.abcde.1234/image/1234/download/ "
            'HTTP/1.1" 200 123456 "https://www.google.com/" '
            '"Mozilla/5.0 (Linux; Android 13; testPhone ABCDEF) '
            "TestPhone/123.45 (KHTML, like Gecko) TestBrowser"
            '/21.0 Chrome/120.3.4567.890 Mobile Safari/537.36"\n'
            "cjs.testing-url222.com:123 34.34.253.34 - - "
            '[02/Jun/2023:04:55:37 +0000] "GET /testing-url/'
            "10.56789/cjs.fghijk/testing/download/ "
            'HTTP/1.1" 200 123456 "https://www.google.com/" '
            '"Mozilla/5.0 (Linux; Android 13; testPhone ABCDEF) '
            "TestPhone/123.45 (KHTML, like Gecko) TestBrowser"
            '/21.0 Chrome/120.3.4567.890 Mobile Safari/537.36"\n'
        )
        self.create_file_log(self.file_log_apache, content)
        filter_groups = []
        for mode in self.modes_cjs:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)
        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with gzip.open(
                self.logs_files + self.file_log_apache,
                "rt",
                encoding="utf-8"
        ) as f:
            result = list(processor.process(f))
        # Assert the result has got the right content
        self.assertEqual(
            result[0][0],
            'tag:testing.eu,2023:readership:abc-html',
        )
        self.assertEqual(
            str(result[0][1]),
            (
                'Request 2023-06-02 04:55:37, 12.34.253.12, '
                'https://abc.test.testing/testing-url/10.3456768/cjs'
                '.abcde.1234/image/1234/download'
            )
        )
        self.assertEqual(
            result[1][0],
            'tag:testing.eu,2023:readership:abc-html',
        )
        self.assertEqual(
            str(result[1][1]),
            (
                'Request 2023-06-02 04:55:37, 34.34.253.34, '
                'https://abc.test.testing/testing-url/10.56789/cjs.'
                'fghijk/testing/download'
            )
        )

    def test_apache_log_start_date_and_end_date_no_match(self) -> None:
        """Test the class LogStream the same way as is executed
        from the plugin"""
        start_date = '2023-05-01'
        end_date = '2023-05-21'
        content = str.encode(
            "cjs.testing-url.com:123 12.90.253.37 - - "
            '[02/Jun/2023:04:55:37 +0000] "GET /testing-url/'
            "10.3456768/cjs.abcde.1234/image/1234/download/ "
            'HTTP/1.1" 200 123456 "https://www.google.com/" '
            '"Mozilla/5.0 (Linux; Android 13; testPhone ABCDEF) '
            "TestPhone/123.45 (KHTML, like Gecko) TestBrowser"
            '/21.0 Chrome/120.3.4567.890 Mobile Safari/537.36"'
        )
        self.create_file_log(self.file_log_apache, content)
        filter_groups = []
        for mode in self.modes_cjs:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)
        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        # Assert the result has got the right content
        with gzip.open(
                self.logs_files + self.file_log_apache,
                "rt",
                encoding="utf-8"
        ) as f:
            result = list(processor.process(f))
        self.assertEqual([], result)

    def test_apache_log_stream_no_match_file_structure_is_ok(self) -> None:
        """Test no match with the current filters even if
        the file structure is fine, will return empty list."""
        start_date = '2023-06-01'
        end_date = '2023-06-03'
        content = str.encode(
            "abcdef.tyestset.com:123 12.34.253.37 - - "
            '[03/Jun/2023:04:51:45 +0000] "GET /test'
            '/test/12.3456/abc.v12a1.1234/ HTTP/1.1" '
            '200 49221 "-" "test/5.0 (Windows NT 10.0; '
            'Win64; x64; ab:123.4) test/20100101 test/112.0"'
        )
        self.create_file_log(self.file_log_apache, content)
        filter_groups = []
        for mode in self.modes_cjs:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)
        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with gzip.open(
                self.logs_files + self.file_log_apache,
                "rt",
                encoding="utf-8"
        ) as f:
            result = list(processor.process(f))
        self.assertEqual([], result)

    def test_apache_log_stream_raises_error_no_matches_lines(self) -> None:
        """Test the class LogStream fails because there is no match.
        The line method will raise exception because the structure of
        the input file is wrong."""
        start_date = '2023-06-02'
        end_date = '2023-06-04'
        content = b'test_worng - wrong structure"'
        self.create_file_log(self.file_log_apache, content)
        filter_groups = []
        for mode in self.modes_cjs:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)

        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )

        with self.assertRaises(AttributeError):
            with gzip.open(
                    self.logs_files + self.file_log_apache,
                    "rt",
                    encoding="utf-8"
            ) as f:
                result = list(processor.process(f))

    def test_apache_log_stream_no_match_request_is_wrong_right_structure(
        self
    ) -> None:
        start_date = '2023-06-03'
        end_date = '2023-06-04'
        """Test the class LogStream there is no match.
        The file structure is right but the request is wrong."""
        content = str.encode(
            "abcdef.presstest.com:123 12.34.253.37 - - "
            '[03/Jun/2023:04:51:45 +0000] "ERROR /articles/'
            'abstract/12.3456/abc.v12a1.1234/ HTTP/1.1" 200 '
            '49221 "-" "Mozilla/5.0 (Windows NT 10.0; Win64; '
            'x64; ab:123.4) Gecko/12345678 Firefox/112.0"'
        )
        self.create_file_log(self.file_log_apache, content)
        filter_groups = []
        for mode in self.modes_cjs:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)

        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with gzip.open(
                self.logs_files + self.file_log_apache,
                "rt",
                encoding="utf-8"
        ) as f:
            result = list(processor.process(f))
        self.assertEqual([], result)

    def test_apache_log_stream_raises_error_there_is_no_timestamp(self) -> None:
        """Test the class LogStream there is no match because of
        missing timestamp."""
        start_date = '2023-06-03'
        end_date = '2023-06-04'
        content = str.encode(
            "abcdef.presstest.com:123 12.12.345.67 - - 'NO TIMESTAMP ' GET "
            "/articles/abstract/12.3456/abcdef.v52i2.8160/ "
            'HTTP/1.1" 200 4522 "-" "Scoop.it"'
        )
        self.create_file_log(self.file_log_apache, content)
        filter_groups = []
        for mode in self.modes_cjs:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)

        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with self.assertRaises(AttributeError):
            with gzip.open(
                    self.logs_files + self.file_log_apache,
                    "rt",
                    encoding="utf-8"
            ) as f:
                result = list(processor.process(f))

    def test_apache_log_stream_wrong_request_and_wrong_structure_raises_error(
        self,
    ) -> None:
        start_date = '2023-06-03'
        end_date = '2023-06-04'
        """Test the class LogStream fails because there is no match.
        The file structure is wrong and the request is wrong."""
        content = str.encode(
            "abcdef.presstest.com:123 146.249.11.3 - - "
            "[02/Jun/2023:08:18:56 +0000] --->ErrorHEADtest "
            '"-" "Scoop.it"'
        )
        self.create_file_log(self.file_log_apache, content)
        filter_groups = []
        for mode in self.modes_cjs:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)

        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with self.assertRaises(AttributeError):
            with gzip.open(
                    self.logs_files + self.file_log_apache,
                    "rt",
                    encoding="utf-8"
            ) as f:
                result = list(processor.process(f))

    def test_apache_log_stream_no_matches_line_r_n_ua_re(self) -> None:
        """Test the class LogStream fails because there is no match.
        The content of the files doesn't match r_n_ua_re regex."""
        start_date = '2023-06-01'
        end_date = '2023-06-02'
        content = str.encode(
            "abcdef.presstest.com:123 12.34.253.37 - - "
            '[01/Jun/2023:06:52:43 +0000] "GET /jms/public'
            '/journals/1/journalFavicon_en_US.ico HTTP/1.1" 404 '
            '5010 "test error!"test error!"'
        )
        self.create_file_log(self.file_log_apache, content)
        filter_groups = []
        for mode in self.modes_cjs:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)

        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with self.assertRaises(AttributeError):
            with gzip.open(
                    self.logs_files + self.file_log_apache,
                    "rt",
                    encoding="utf-8"
            ) as f:
                result = list(processor.process(f))

    def test_apache_log_stream_ignores_reuqest_if_its_missing_parts(
        self
    ) -> None:
        """Test the class LogStream ignores urls that have a missing method,
        url or version."""
        start_date = '2023-06-01'
        end_date = '2023-06-02'
        content = str.encode(
            "abc.presstest.com:123 12.34.253.37 - - "
            '[01/Jun/2023:06:52:43 +0000] "GET /jms/public'
            '/journals/1/journalFavicon_en_US.ico" 404 5010 '
            '"https://abc.defg.info/'
            'abc.v52i2.8023/" "Mozilla/5.0 (Linux; Android 10; K) '
            "Appltest_log_stream_ignores_reuqest_if_its_missing_parts"
            "eWebKit/537.36 (KHTML, like Gecko) Chrome/"
            '113.0.0.0 Mobile Safari/537.36"abc.presstest.'
            "com:123 12.12.345.67 - - [01/Jun/2023:06:52:50 +0000] "
            '"GET /jms/public/journals/1/journalFavicon_en_US.ico" '
        )
        self.create_file_log(self.file_log_apache, content)
        filter_groups = []
        for mode in self.modes_cjs:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)

        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with gzip.open(
                self.logs_files + self.file_log_apache,
                "rt",
                encoding="utf-8"
        ) as f:
            result = list(processor.process(f))
        self.assertEqual([], result)

    def test_apache_log_stream_the_ip_address_is_wrong_no_match(
        self,
    ) -> None:
        """Test the class LogStream fails because there is no match.
        The file structure is wrong and the request is wrong."""
        start_date = '2023-06-01'
        end_date = '2023-06-02'
        content = str.encode(
            "cjs.testing-url.com:123 99.999.999.999 - - "
            '[02/Jun/2023:04:55:37 +0000] "GET /testing-url/'
            "10.3456768/cjs.abcde.1234/image/1234/download/ "
            'HTTP/1.1" 200 123456 "https://www.google.com/" '
            '"Mozilla/5.0 (Linux; Android 13; testPhone ABCDEF) '
            "TestPhone/123.45 (KHTML, like Gecko) TestBrowser"
            '/21.0 Chrome/120.3.4567.890 Mobile Safari/537.36"'
        )
        self.create_file_log(self.file_log_apache, content)
        filter_groups = []
        for mode in self.modes_cjs:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)

        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with gzip.open(
                self.logs_files + self.file_log_apache,
                "rt",
                encoding="utf-8"
        ) as f:
            result = list(processor.process(f))
        self.assertEqual([], result)

    def test_apache_logs_no_matches_start_date_wrong_file(self):
        """The search date is in august and the file name is in June."""
        start_date = '2023-08-01'
        end_date = '2023-08-02'
        content = str.encode(
            "cjs.testing-url.com:123 12.90.253.37 - - "
            '[02/Jun/2023:04:55:37 +0000] "GET /testing-url/'
            "10.3456768/cjs.abcde.1234/image/1234/download/ "
            'HTTP/1.1" 200 123456 "https://www.google.com/" '
            '"Mozilla/5.0 (Linux; Android 13; testPhone ABCDEF) '
            "TestPhone/123.45 (KHTML, like Gecko) TestBrowser"
            '/21.0 Chrome/120.3.4567.890 Mobile Safari/537.36"'
        )
        self.create_file_log(self.file_log_apache, content)
        filter_groups = []
        for mode in self.modes_nginx:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)

        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with gzip.open(
                self.logs_files + self.file_log_apache,
                "rt",
                encoding="utf-8"
        ) as f:
            result = list(processor.process(f))
        self.assertEqual([], result)

    def test_nginx_logs_is_successful_two_matches(self):
        """Test the nginx logs are processed as well
        the four lines will be matched."""
        start_date = '2023-07-05'
        end_date = '2023-07-06'
        content = str.encode(
            '25.123.123.1 - - [06/Jul/2023:15:04:34 +0000] "GET '
            '/test.txt HTTP/1.0" 302 555 "-" '
            '"Mozitest (http://Moziddler.io/about)"\n'
            '25.123.123.2 - - [06/Jul/2023:15:49:29 +0000] "GET '
            '/test/books/e/10.5334/bbc HTTP/1.0" 200 547 "-" '
            '"Ozitest/7.7 (X11; Xunil x86_64; rh:123.1) SalaTest/20100101 "'
            "prüfen/123.1\n"
            '25.123.123.3 - - [06/Jul/2023:20:15:19 +0000] "GET '
            '/test-query HTTP/1.0" 404 502 "-" '
            '"Mozitest (http://Moziddler.io/about)"\n'
            '25.123.123.4 - - [06/Jul/2023:21:23:18 +0000] "GET '
            '/test/10.5334/bay HTTP/1.0" 200 496 "-" '
            '"Ozitest/7.7 (X11; Xunil x86_64; rh:123.1) SalaTest/20100101 '
            'prüfen/123.1"'
        )
        self.create_file_log(self.file_log_nginx, content)
        filter_groups = []
        for mode in self.modes_nginx:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)

        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with gzip.open(
                self.logs_files + self.file_log_nginx,
                "rt",
                encoding="utf-8"
        ) as f:
            result = list(processor.process(f))
        self.assertEqual(
            result[0][0],
            'https://testing.test-eu.org/abc/v1'
        )
        self.assertEqual(
            str(result[0][1]),
            (
                'Request 2023-07-06 15:49:29, 25.123.123.2, '
                'https://abc.test.testing/test/books/e/10.5334/bbc'
            )
        )
        self.assertEqual(
            result[1][0],
            'https://testing.test-eu.org/abc/v1'
        )
        self.assertEqual(
            str(result[1][1]),
            (
                'Request 2023-07-06 21:23:18, 25.123.123.4, '
                'https://abc.test.testing/test/10.5334/bay'
            )
        )

    def test_nginx_logs_no_matches_line_r_n_ua_re(self):
        """Test the nginx logs are processed as well
        the four lines will be matched."""
        start_date = '2023-07-05'
        end_date = '2023-07-07'
        content = str.encode(
            '25.123.123.1 - - [06/Jul/2023:15:04:34 +0000] "GET '
            '/test.txt HTTP/1.0" 302 555 "-" '
            '"WrongTESTUSERAGENT"\n'
        )
        self.create_file_log(self.file_log_nginx, content)
        filter_groups = []
        for mode in self.modes_nginx:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)

        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with gzip.open(
                self.logs_files + self.file_log_nginx,
                "rt",
                encoding="utf-8"
        ) as f:
            result = list(processor.process(f))
        self.assertEqual([], result)

    def test_nginx_logs_no_matches_start_date_wrong_file(self):
        """The search date is in august and the file name is in July."""
        start_date = '2023-07-31'
        end_date = '2023-08-02'
        content = str.encode(
            '25.123.123.1 - - [06/Jul/2023:15:04:34 +0000] "GET '
            '/test.txt HTTP/1.0" 302 555 "-" '
            '"Mozitest (http://Moziddler.io/about)"\n'
        )
        self.create_file_log(self.file_log_nginx, content)
        filter_groups = []
        for mode in self.modes_nginx:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)

        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with gzip.open(
                self.logs_files + self.file_log_nginx,
                "rt",
                encoding="utf-8"
        ) as f:
            result = list(processor.process(f))
        self.assertEqual([], result)

    def test_nginx_logs_no_matches_start_date_wrong_line(self):
        """The search date is in July and the line is in august."""
        start_date = '2023-07-06'
        end_date = '2023-07-07'
        content = str.encode(
            '25.123.123.1 - - [06/Aug/2023:15:04:34 +0000] "GET '
            '/test.txt HTTP/1.0" 302 555 "-" '
            '"Mozitest (http://Moziddler.io/about)"\n'
        )
        self.create_file_log(self.file_log_nginx, content)
        filter_groups = []
        for mode in self.modes_nginx:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)

        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with gzip.open(
                self.logs_files + self.file_log_nginx,
                "rt",
                encoding="utf-8"
        ) as f:
            result = list(processor.process(f))
        self.assertEqual([], result)

    def test_nginx_log_stream_no_match_file_structure_is_ok(self) -> None:
        """Test the nginx logs are processed as well
        the four lines will be matched."""
        start_date = '2023-07-06'
        end_date = '2023-07-07'
        content = str.encode(
            '25.123.123.1 - - [06/Jul/2023:15:04:34 +0000] "GET '
            'NOMATCH/test.txt HTTP/1.0" 302 555 "-" '
            '"Mozitest (http://Moziddler.io/about)"\n'
            '25.123.123.2 - - [06/Jul/2023:15:49:29 +0000] "GET '
            '/NOMATCH/books/e/10.5334/bbc HTTP/1.0" 200 547 "-" '
            '"Ozitest/7.7 (X11; Xunil x86_64; rh:123.1) SalaTest/20100101 "'
            "prüfen/123.1\n"
        )
        self.create_file_log(self.file_log_nginx, content)
        filter_groups = []
        for mode in self.modes_nginx:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)

        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with gzip.open(
                self.logs_files + self.file_log_nginx,
                "rt",
                encoding="utf-8"
        ) as f:
            result = list(processor.process(f))
        self.assertEqual([], result)

    def test_nginx_log_stream_no_match_request_is_wrong_right_structure(
        self
    ) -> None:
        """Test the class LogStream there is no match.
        The file structure is right but the request is wrong."""
        start_date = '2023-07-06'
        end_date = '2023-07-07'
        content = str.encode(
            '25.123.123.3 - - [06/Jul/2023:20:15:19 +0000] "ERROR '
            '/test-query HTTP/1.0" 404 502 "-" '
            '"Mozitest (http://Moziddler.io/about)"\n'
            '25.123.123.4 - - [06/Jul/2023:21:23:18 +0000] "WRONG REQUEST '
            '/test/10.5334/bay HTTP/1.0" 200 496 "-" '
            '"Ozitest/7.7 (X11; Xunil x86_64; rh:123.1) SalaTest/20100101 '
            'prüfen/123.1"'
        )
        self.create_file_log(self.file_log_nginx, content)
        filter_groups = []
        for mode in self.modes_nginx:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)

        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with gzip.open(
                self.logs_files + self.file_log_nginx,
                "rt",
                encoding="utf-8"
        ) as f:
            result = list(processor.process(f))
        self.assertEqual([], result)

    def test_nginx_log_stream_raises_error_there_is_no_timestamp(self) -> None:
        """Test the class LogStream there is no match because of
        missing timestamp."""
        start_date = '2023-07-06'
        end_date = '2023-07-07'
        content = str.encode(
            '25.123.123.3 - - NO TIMESTAMPP "GET '
            '/test-query HTTP/1.0" 404 502 "-" '
            '"Mozitest (http://Moziddler.io/about)"'
        )
        self.create_file_log(self.file_log_nginx, content)
        filter_groups = []
        for mode in self.modes_nginx:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)

        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with self.assertRaises(AttributeError):
            with gzip.open(
                    self.logs_files + self.file_log_nginx,
                    "rt",
                    encoding="utf-8"
            ) as f:
                result = list(processor.process(f))

    def test_nginx_log_stream_no_match_wrong_request_and_structure_raises_error(
        self,
    ) -> None:
        start_date = '2023-07-06'
        end_date = '2023-07-07'
        """Test the class LogStream fails because there is no match.
        The file structure is wrong and the request is wrong."""
        content = str.encode(
            '146.249.11.159 - - [06/Jul/2023:21:23:18 +0000] ----> ERRORRR'
            '/wrong/10.5334/bay HTTP/1.0" 200 496 "-" '
            '"Ozitest/7.7 (X11; Xunil x86_64; rh:123.1) SalaTest/20100101 '
            'prüfen/123.1"'
        )
        self.create_file_log(self.file_log_nginx, content)
        filter_groups = []
        for mode in self.modes_cjs:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)

        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with self.assertRaises(AttributeError):
            with gzip.open(
                    self.logs_files + self.file_log_nginx,
                    "rt",
                    encoding="utf-8"
            ) as f:
                result = list(processor.process(f))

    def test_nginx_log_stream_the_ip_address_is_wrong_will_raise_error(
        self,
    ) -> None:
        start_date = '2023-07-05'
        end_date = '2023-07-07'
        """Test the class LogStream fails because there is no match.
        The file structure is wrong and the request is wrong."""
        content = str.encode(
            '999.999.999.999 - - [06/Jul/2023:15:04:34 +0000] "GET '
            '/test.txt HTTP/1.0" 302 555 "-" '
            '"Mozitest (http://Moziddler.io/about)"'
        )
        self.create_file_log(self.file_log_nginx, content)
        filter_groups = []
        for mode in self.modes_cjs:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)

        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with gzip.open(
                self.logs_files + self.file_log_nginx,
                "rt",
                encoding="utf-8"
        ) as f:
            result = list(processor.process(f))
        self.assertEqual([], result)

    def test_nginx_log_stream_the_start_date_and_end_date_no_match(
        self,
    ) -> None:
        start_date = '2023-07-01'
        end_date = '2023-07-02'
        """Test the class LogStream fails because there is no match.
        The file structure is wrong and the request is wrong."""
        content = str.encode(
            '25.123.123.1 - - [06/Jul/2023:15:04:34 +0000] "GET '
            '/test.txt HTTP/1.0" 302 555 "-" '
            '"Mozitest (http://Moziddler.io/about)"\n'
            '25.123.123.2 - - [06/Jul/2023:15:49:29 +0000] "GET '
            '/test/books/e/10.5334/bbc HTTP/1.0" 200 547 "-" '
            '"Ozitest/7.7 (X11; Xunil x86_64; rh:123.1) SalaTest/20100101 "'
            "prüfen/123.1\n"
        )
        self.create_file_log(self.file_log_nginx, content)
        filter_groups = []
        for mode in self.modes_cjs:
            filters = (
                mode.get("measure"),
                make_filters(
                    mode.get("regex"),
                    ["8.8.8.8", "9.9.9.9"],
                ),
                mode.get("regex")
            )
            filter_groups.append(filters)

        processor = LogProcessor(
            filter_groups,
            self.url_prefix,
            start_date,
            end_date
        )
        with gzip.open(
                self.logs_files + self.file_log_nginx,
                "rt",
                encoding="utf-8"
        ) as f:
            result = list(processor.process(f))
        self.assertEqual([], result)
